#include <display_syscalls.h>
#include <keyboard_syscalls.h>
#include <keyboard.hpp>
#include <color.h>
#include <disp_tools.hpp>
#include <display.h>
#include <HEAP_syscalls.h>
#include <RTC_syscalls.h>
#include <CONVERT_syscalls.h>
#include <stdlib.h>
#include <sprite.h>
#include <MCS_syscalls.h>
#include "Minesweeper2.h"
//
#define MAX_WIDTH	19
#define MAX_HEIGHT	8

#define MIN(a, b)	a < b ? a : b
#define MAX(a, b)	a > b ? a : b

#define FILLED -1
#define FLAGGED -2
#define QUESTION -3

int numMines = 8;
int fieldWidth = 8;
int fieldHeight = 8;

int maxMines = MAX_WIDTH*MAX_HEIGHT-1;
int minMines = 5;
int maxWidth = MAX_WIDTH;
int minWidth = 5;
int maxHeight = MAX_HEIGHT;
int minHeight = 5;

int xOff = 0;
int yOff = 0;

int total = 0;
int numFlags = 0;

char* rightArrow = "  \xE5\xE9";
	
char visibleField[MAX_WIDTH][MAX_HEIGHT];
char mineField[MAX_WIDTH][MAX_HEIGHT];

// This feels dangerous.
#define DIRNAME (unsigned char*)"@MINES"
#define STATEFILE_V (unsigned char*)"StateV"
#define STATEFILE_M (unsigned char*)"StateM"
#define STATEFILE_S (unsigned char*)"StateS"
#define SETTINGSFILE (unsigned char*)"Set"

const unsigned int default_fkeys[] = 
  { 0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0 };

int main(void) {
	Set_FKeys1(0, (unsigned int*)default_fkeys);
	Set_FKeys2(0);
	Bdisp_EnableColor(1);
	MCS_CreateDirectory(DIRNAME);

	int ticks = RTC_GetTicks();
	srand(ticks);
	
	DefineStatusAreaFlags(DSA_SETDEFAULT, 0, 0, 0);	// necessary, if the status area should be used by system functions!
	DefineStatusAreaFlags(3, SAF_BATTERY|SAF_TEXT|SAF_GLYPH, 0, 0);
	
	int x;
	int y;
	int key;

	LoadSettings(SETTINGSFILE);
start:
	total = 0;
	x = 0;
	y = 0;
	numFlags = 0;
	CreateFields(0, 0);

	while (1) {
		Bdisp_AllCr_VRAM();
		Bdisp_EnableColor(1);
		
		unsigned char out[3];
		itoa(numMines-numFlags, out);
		DefineStatusMessage(concat("  Minesweeper!    Mines: ", (char*)out), 1, 0, 0);
		DisplayStatusArea();
		SetBottomText("Click Flag Save  Load Props New");
		
		DrawField();
		//VRAM_XORSprite(cursor, x*20, y*20, Tile_Width, Tile_Height);
		DrawTile((color_t*)cursor, x, y, 1);
		GetKey(&key);
		switch (key) {
			case KEY_CTRL_UP:
				y = y == 0 ? fieldHeight-1 : y-1;
				break;
			case KEY_CTRL_DOWN:
				y = y == fieldHeight-1 ? 0 : y+1;
				break;
			case KEY_CTRL_LEFT:
				x = x == 0 ? fieldWidth-1 : x-1;
				break;
			case KEY_CTRL_RIGHT:
				x = x == fieldWidth-1 ? 0 : x+1;
				break;
			case KEY_CTRL_F2:
				if (visibleField[x][y] == FILLED) {
					visibleField[x][y] = FLAGGED;
					numFlags++;
				} else if (visibleField[x][y] == FLAGGED) {
					visibleField[x][y] = QUESTION;
					numFlags--;
				} else if (visibleField[x][y] == QUESTION) {
					visibleField[x][y] = FILLED;
				}
				break;
			case KEY_CTRL_F1:
				if (visibleField[x][y] == FILLED || visibleField[x][y] == QUESTION) {
					int num = Cascade(x, y, total);
					if (!num) {
						Die(x, y);
						goto start;
					}
					total += num;
					if (total == fieldHeight*fieldWidth - numMines) {
						Win();
						goto start;
					}
				}
				break;
			case KEY_CTRL_F3:
				SaveState();
				break;
			case KEY_CTRL_F4:
				LoadState();
				break;
			case KEY_CTRL_F5:
				if (Settings()) {
					SaveSettings(SETTINGSFILE);
					goto start;
				}
				break;
			case KEY_CTRL_F6:
				goto start;
				break;
		}
	}

	return 0;
}

void About() {
	Bdisp_AllCr_VRAM();
	DefineStatusMessage("  Minesweeper v2.1", 1, 0, 0);
	DisplayStatusArea();
	PrintXY(1, 1, "   Merthsoft Creations ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 2, "   (Shaun McFall 2012) ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 3, "         Thanks:       ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 4, "      Andrew - Art     ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 5, "  Fishbot - Prizm Stuff", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 6, "     KermM - C Stuff   ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 7, "     Sarah - Nagging   ", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	SetBottomText("(Press any key)");
	int key;
	do {
		GetKey(&key);
	} while (!key);
}

int Settings() {
	int key;
	int y = 1;
	unsigned char out[3];

	int newValues[] = {numMines, fieldWidth, fieldHeight};
	int maxValues[] = {maxMines, maxWidth, maxHeight};
	int minValues[] = {minMines, minWidth, minHeight};

	while (1) {
		Bdisp_AllCr_VRAM();
		DefineStatusMessage("  Minesweeper!   Settings.", 1, 0, 0);
		DisplayStatusArea();

		itoa(newValues[0], out);
		PrintXY(2, 1,  concat("  Number of Mines: ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
		itoa(newValues[1], out);
		PrintXY(2, 2,  concat("  Field Width:     ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
		itoa(newValues[2], out);
		PrintXY(2, 3,  concat("  Field Height:    ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

		SetBottomText("Back      About             New");

		PrintXY(1, y, rightArrow, TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

		GetKey(&key);
		switch (key) {
			case KEY_CTRL_UP:
				y = y == 1 ? 3 : y-1;
				break;
			case KEY_CTRL_DOWN:
				y = y == 3 ? 1 : y+1;
				break;
			case KEY_CTRL_LEFT:
				newValues[y-1] = MAX(newValues[y-1]-1, minValues[y-1]);
				// Max mines is total size - 1
				maxValues[0] = newValues[1] * newValues[2] - 1;
				newValues[0] = MIN(newValues[0], maxValues[0]);
				break;
			case KEY_CTRL_RIGHT:
				newValues[y-1] = MIN(newValues[y-1]+1, maxValues[y-1]);
				// Max mines is total size - 1
				maxValues[0] = newValues[1] * newValues[2] - 1;
				newValues[0] = MIN(newValues[0], maxValues[0]);
				break;
			case KEY_CTRL_F1:
				return 0;
				break;
			case KEY_CTRL_F3:
				About();
				break;
			case KEY_CTRL_F6:
				numMines = newValues[0];
				fieldWidth = newValues[1];
				fieldHeight = newValues[2];
				return 1;
				break;
		}
	}
}

char* concat(char* str1, char* str2) {
	int len1 = 0;
	int len2 = 0;
	for (;str1[len1] != '\0'; len1++);
	for (;str2[len2] != '\0'; len2++);

	char* str3 = (char*)malloc(len1 + len2 + 1);
	int i;
	for (i = 0; i < len1; str3[i] = str1[i], i++);
	for (i = 0; i < len2; str3[len1+i] = str2[i], i++);
	str3[len1 + len2] = 0;
	
	return str3;
}

void PrintXYVisibleSpot(int i, int j) {
	DrawTile(getVisibleType(i,j), i, j, 0);
}

void DrawTile(color_t* tile, int i, int j, int xor) {
	if (xor) {
		VRAM_XORSprite(tile, i*20+xOff, j*20+yOff, Tile_Width, Tile_Height);
	} else {
		VRAM_CopySprite(tile, i*20+xOff, j*20+yOff, Tile_Width, Tile_Height);
	}
}

void DrawField() {
	int i, j;
	// Inefficient? Sure. But I don't care.
	xOff = LCD_WIDTH_PX/2 - (20*fieldWidth)/2;
	yOff = LCD_HEIGHT_PX/2 - (20*fieldHeight)/2;
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			PrintXYVisibleSpot(i, j);
		}
	}
}

void Win() {
	int i, j;
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			PrintXYVisibleSpot(i, j);
			if (mineField[i][j]) {
				//PrintXY(i, j, star, TEXT_MODE_INVERT, TEXT_COLOR_YELLOW);
				DrawTile((color_t*)win, i, j, 0);
			}
		}
	}
	DefineStatusMessage("  Minesweeper!    You win! :D", 1, 0, 0);
	DisplayStatusArea();
	SetBottomText("                            New");

	int key;
	do {
		GetKey(&key);
	} while (key != KEY_CTRL_F6);
}

void Die(int cursorX, int cursorY) {
	int i, j;
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			//PrintXY(i, j, getVisibleType(i-1,j-1), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
			PrintXYVisibleSpot(i, j);
			if (mineField[i][j]) {
				//PrintXY(i, j, mine, visibleField[i-1][j-1] == FLAGGED?TEXT_MODE_INVERT:TEXT_MODE_NORMAL, TEXT_COLOR_RED);
				if (i == cursorX && j == cursorY) {
					DrawTile((color_t*)explode, i, j, 0);
				} else {
					DrawTile(visibleField[i][j] == FLAGGED?(color_t*)flagged:(color_t*)mine, i, j, 0);
				}
			} else if (visibleField[i][j] == FLAGGED) {
				DrawTile((color_t*)misflagged, i, j, 0);
			}
		}
	}
	DefineStatusMessage("  Minesweeper!    You lose! D:", 1, 0, 0);
	DisplayStatusArea();
	SetBottomText("                            New");

	int key;
	do {
		GetKey(&key);
	} while (key != KEY_CTRL_F6);
}

int Cascade(int x, int y, int total) {
	if (total == 0) {
		CreateFields(x, y);
	}
	if (x < 0 || y < 0 || x >= fieldWidth || y >= fieldHeight)
      return 0;
    if (visibleField[x][y] >= 0 || visibleField[x][y] == -2)
      return 0;
    if (mineField[x][y] == 1)
      return 0;
    int neighbors = 0;
	int i, j;
    for (i = -1; i <= 1; i++) {
      for (j = -1; j <= 1; j++) {
        if (!(x+i < 0 || y+j < 0 || x+i >= fieldWidth || y+j >= fieldHeight))
          neighbors += mineField[x+i][y+j];
      }
    }
    visibleField[x][y] = neighbors;
    int num = 1;
    if (neighbors == 0) {
      for (i = -1; i <= 1; i++) {
        for (j = -1; j <= 1; j++) {
          num += Cascade(x + i, y + j, num);
        }
      }
    }
    return num;
}

void CreateFields(int x, int y) {
	int i, j;
	
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			visibleField[i][j] = FILLED;
			mineField[i][j] = 0;
		}
	}

	for (i = 0; i < numMines; i++) {
		int mineX, mineY;
		do {
			mineX = rand() % fieldWidth;
			mineY = rand() % fieldHeight;
		} while (mineField[mineX][mineY] != 0 || (mineX == x && mineY == y));
		mineField[mineX][mineY] = 1;
	}
}

color_t* getVisibleType(int i, int j) {
	const color_t* type = 0;
	switch (visibleField[i][j]) {
		case FILLED:
			type = filled;
			break;
		case FLAGGED:
			type = flagged;
			break;
		case QUESTION:
			type = question;
			break;
		case 0:
			type = zero;
			break;
		case 1:
			type = one;
			break;
		case 2:
			type = two;
			break;
		case 3:
			type = three;
			break;
		case 4:
			type = four;
			break;
		case 5:
			type = five;
			break;
		case 6:
			type = six;
			break;
		case 7:
			type = seven;
			break;
		case 8:
			type = eight;
			break;
	}
	return (color_t*)type;
}

void SetBottomText(char* text) {
	int i,j;
	for (i = 198; i < LCD_HEIGHT_PX; i++) {
		for (j = 0; j < 380; j++) {
			VRAM_Plot(j, i, COLOR_BLACK);
		}
	}

	PrintMiniFix(0, 175, text, 0, COLOR_WHITE, COLOR_BLACK);
}

int LoadState() {
	int size;
	if (MCSGetDlen2(DIRNAME, STATEFILE_V, &size)) { return 1; }
	if (size == 0) { return 5; }
	if (MCSGetData1(0, MAX_HEIGHT*MAX_WIDTH, visibleField)) { return 2; }
	if (MCSGetDlen2(DIRNAME, STATEFILE_M, &size)) { return 3; }
	if (size == 0) { return 6; }
	if (MCSGetData1(0, MAX_HEIGHT*MAX_WIDTH, mineField)) { return 4; }

	numFlags = 0;
	total = 0;

	LoadSettings(STATEFILE_S);

	int i, j;
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			if (visibleField[i][j] == FLAGGED) {
				numFlags++;
			} else if (visibleField[i][j] != FILLED) {
				total++;
			}
		}
	}

	return 0;
}

int SaveState() {
	if (MCSDelVar2(DIRNAME, STATEFILE_V)) { /*return 1;*/ }
	if (MCSPutVar2(DIRNAME, STATEFILE_V, MAX_HEIGHT*MAX_WIDTH, visibleField)) { return 2; }
	if (MCSDelVar2(DIRNAME, STATEFILE_M)) { /*return 3;*/ }
	if (MCSPutVar2(DIRNAME, STATEFILE_M, MAX_HEIGHT*MAX_WIDTH, mineField)) { return 4; }

	SaveSettings(STATEFILE_S);
	
	return 0;
}

void LoadSettings(unsigned char* file) {
	int size;
	MCSGetDlen2(DIRNAME, file, &size);
	if (size == 0) { return; }
	
	char buffer[3];
	MCSGetData1(0, 3, buffer);
	fieldWidth = buffer[0];
	fieldHeight = buffer[1];
	numMines = (unsigned char)buffer[2];
}

void SaveSettings(unsigned char* file) {
	MCSDelVar2(DIRNAME, file);
	char buffer[3] = {fieldWidth, fieldHeight, numMines};
	MCSPutVar2(DIRNAME, file, 3, buffer);
}

int PrintMiniFix(int x, int y, const char*Msg, const int flags, const short color, const short bcolor) {
	int i = 0, dx;
	unsigned short width;
	void*p;

	while (Msg[i]){
		p = GetMiniGlyphPtr( Msg[ i ], &width );
		dx = ( 12 - width ) / 2;
		if ( dx > 0 ) {
			PrintMiniGlyph( x, y, (void*)empty, flags, dx, 0, 0, 0, 0, color, bcolor, 0 );
		}else dx = 0;
		PrintMiniGlyph( x+dx, y, p, flags, width, 0, 0, 0, 0, color, bcolor, 0 );
		if ( width+dx < 12 ){
			PrintMiniGlyph( x+width+dx, y, (void*)empty, flags, 12-width-dx, 0, 0, 0, 0, color, bcolor, 0 );
		}
		x += 12;
		i++;
	}
	return x;
}